# -*- coding: utf-8 -*-
"""
A sober python base library
"""
# SPDX-License-Identifier: LGPL-3.0

__version__ = "0.8.6"

from .util import *   # noqa: F403, F401
from .obj  import *   # noqa: F403, F401
